@@ -21,23 +21,23 @@
 
 #include <QDateTime>
 #include <QHash>
-#include <QObject>
+#include <QMutex>
 #include <QPointer>
 #include <QScopedPointer>
+#include <QTimer>
 
 #include "config-keepassx.h"
 #include "crypto/kdf/AesKdf.h"
 #include "crypto/kdf/Kdf.h"
 #include "format/KeePass2.h"
-#include "keys/PasswordKey.h"
 #include "keys/CompositeKey.h"
+#include "keys/PasswordKey.h"
 
 class Entry;
 enum class EntryReferenceType;
 class FileWatcher;
 class Group;
 class Metadata;
-class QTimer;
 class QIODevice;
 
 struct DeletedObject
@@ -81,11 +81,11 @@ public:
     void releaseData();
 
     bool isInitialized() const;
-    void setInitialized(bool initialized);
-    bool isModified() const;
+    bool isModified(bool includeNonDataChanges = false) const;
     void setEmitModified(bool value);
     bool isReadOnly() const;
     void setReadOnly(bool readOnly);
+    bool isSaving();
 
     QUuid uuid() const;
     QString filePath() const;
@@ -114,15 +114,14 @@ public:
 
     QList<QString> commonUsernames();
 
-    bool hasKey() const;
     QSharedPointer<const CompositeKey> key() const;
     bool setKey(const QSharedPointer<const CompositeKey>& key,
                 bool updateChangedTime = true,
                 bool updateTransformSalt = false,
                 bool transformKey = true);
+    QString keyError();
     QByteArray challengeResponseKey() const;
     bool challengeMasterSeed(const QByteArray& masterSeed);
-    bool verifyKey(const QSharedPointer<CompositeKey>& key) const;
     const QUuid& cipher() const;
     void setCipher(const QUuid& cipher);
     Database::CompressionAlgorithm compressionAlgorithm() const;
@@ -131,7 +130,7 @@ public:
     QSharedPointer<Kdf> kdf() const;
     void setKdf(QSharedPointer<Kdf> kdf);
     bool changeKdf(const QSharedPointer<Kdf>& kdf);
-    QByteArray transformedMasterKey() const;
+    QByteArray transformedDatabaseKey() const;
 
     static Database* databaseByUuid(const QUuid& uuid);
 
@@ -139,6 +138,7 @@ public slots:
     void markAsModified();
     void markAsClean();
     void updateCommonUsernames(int topN = 10);
+    void markNonDataChange();
 
 signals:
     void filePathChanged(const QString& oldPath, const QString& newPath);
@@ -155,9 +155,6 @@ signals:
     void databaseDiscarded();
     void databaseFileChanged();
 
-private slots:
-    void startModifiedTimer();
-
 private:
     struct DatabaseData
     {
@@ -167,10 +164,9 @@ private:
         CompressionAlgorithm compressionAlgorithm = CompressionGZip;
 
         QScopedPointer<PasswordKey> masterSeed;
-        QScopedPointer<PasswordKey> transformedMasterKey;
+        QScopedPointer<PasswordKey> transformedDatabaseKey;
         QScopedPointer<PasswordKey> challengeResponseKey;
 
-        bool hasKey = false;
         QSharedPointer<const CompositeKey> key;
         QSharedPointer<Kdf> kdf = QSharedPointer<AesKdf>::create(true);
 
@@ -178,7 +174,7 @@ private:
 
         DatabaseData()
             : masterSeed(new PasswordKey())
-            , transformedMasterKey(new PasswordKey())
+            , transformedDatabaseKey(new PasswordKey())
             , challengeResponseKey(new PasswordKey())
         {
             kdf->randomizeSeed();
@@ -189,10 +185,9 @@ private:
             filePath.clear();
 
             masterSeed.reset();
-            transformedMasterKey.reset();
+            transformedDatabaseKey.reset();
             challengeResponseKey.reset();
 
-            hasKey = false;
             key.reset();
             kdf.reset();
 
@@ -211,11 +206,13 @@ private:
     DatabaseData m_data;
     QPointer<Group> m_rootGroup;
     QList<DeletedObject> m_deletedObjects;
-    QPointer<QTimer> m_timer;
+    QTimer m_modifiedTimer;
+    QMutex m_saveMutex;
     QPointer<FileWatcher> m_fileWatcher;
-    bool m_initialized = false;
     bool m_modified = false;
     bool m_emitModified;
+    bool m_hasNonDataChange = false;
+    QString m_keyError;
 
     QList<QString> m_commonUsernames;
 
