@@ -21,17 +21,38 @@
 
 #include "gui/KMessageWidget.h"
 
+class QTimer;
+
 class MessageWidget : public KMessageWidget
 {
     Q_OBJECT
 
 public:
-    explicit MessageWidget(QWidget* parent = 0);
+    explicit MessageWidget(QWidget* parent = nullptr);
+
+    int autoHideTimeout() const;
+
+    static const int DefaultAutoHideTimeout;
+    static const int LongAutoHideTimeout;
+    static const int DisableAutoHide;
+
+    void setAnimate(bool state);
+
+signals:
+    void showAnimationStarted();
+    void hideAnimationStarted();
 
 public slots:
     void showMessage(const QString& text, MessageWidget::MessageType type);
+    void showMessage(const QString& text, MessageWidget::MessageType type, int autoHideTimeout);
     void hideMessage();
+    void setAutoHideTimeout(int autoHideTimeout);
+    static void openHttpUrl(QString const& url);
 
+private:
+    QTimer* m_autoHideTimer;
+    int m_autoHideTimeout;
+    bool m_animate;
 };
 
 #endif // MESSAGEWIDGET_H
