@@ -17,30 +17,30 @@
 
 #include "EntryModel.h"
 
-#include <QFont>
-#include <QFontMetrics>
-#include <QMimeData>
-#include <QPalette>
 #include <QDateTime>
+#include <QFont>
+#include <QMimeData>
 #include <QPainter>
+#include <QPalette>
 
+#include "core/Config.h"
 #include "core/DatabaseIcons.h"
 #include "core/Entry.h"
 #include "core/Global.h"
 #include "core/Group.h"
 #include "core/Metadata.h"
-
-// String being displayed when hiding content
-const QString EntryModel::HiddenContentDisplay(QString("\u25cf").repeated(6));
-
-// Format used to display dates
-const Qt::DateFormat EntryModel::DateFormat = Qt::DefaultLocaleShortDate;
+#include "core/Resources.h"
+#ifdef Q_OS_MACOS
+#include "gui/osutils/macutils/MacUtils.h"
+#endif
 
 EntryModel::EntryModel(QObject* parent)
     : QAbstractTableModel(parent)
     , m_group(nullptr)
     , m_hideUsernames(false)
     , m_hidePasswords(true)
+    , HiddenContentDisplay(QString("\u25cf").repeated(6))
+    , DateFormat(Qt::DefaultLocaleShortDate)
 {
 }
 
@@ -75,10 +75,9 @@ void EntryModel::setGroup(Group* group)
     makeConnections(group);
 
     endResetModel();
-    emit switchedToListMode();
 }
 
-void EntryModel::setEntryList(const QList<Entry*>& entries)
+void EntryModel::setEntries(const QList<Entry*>& entries)
 {
     beginResetModel();
 
@@ -112,7 +111,6 @@ void EntryModel::setEntryList(const QList<Entry*>& entries)
     }
 
     endResetModel();
-    emit switchedToSearchMode();
 }
 
 int EntryModel::rowCount(const QModelIndex& parent) const
@@ -131,7 +129,7 @@ int EntryModel::columnCount(const QModelIndex& parent) const
         return 0;
     }
 
-    return 12;
+    return 14;
 }
 
 QVariant EntryModel::data(const QModelIndex& index, int role) const
@@ -176,6 +174,9 @@ QVariant EntryModel::data(const QModelIndex& index, int role) const
             if (attr->isReference(EntryAttributes::PasswordKey)) {
                 result.prepend(tr("Ref: ", "Reference abbreviation"));
             }
+            if (entry->password().isEmpty() && !config()->get(Config::Security_PasswordEmptyPlaceholder).toBool()) {
+                result = "";
+            }
             return result;
         case Url:
             result = entry->resolveMultiplePlaceholders(entry->displayUrl());
@@ -184,15 +185,23 @@ QVariant EntryModel::data(const QModelIndex& index, int role) const
             }
             return result;
         case Notes:
-            // Display only first line of notes in simplified format
-            result = entry->notes().section("\n", 0, 0).simplified();
-            if (attr->isReference(EntryAttributes::NotesKey)) {
-                result.prepend(tr("Ref: ", "Reference abbreviation"));
+            if (!entry->notes().isEmpty()) {
+                if (config()->get(Config::Security_HideNotes).toBool()) {
+                    result = EntryModel::HiddenContentDisplay;
+                } else {
+                    // Display only first line of notes in simplified format if not hidden
+                    result = entry->notes().section("\n", 0, 0).simplified();
+                }
+                if (attr->isReference(EntryAttributes::NotesKey)) {
+                    result.prepend(tr("Ref: ", "Reference abbreviation"));
+                }
             }
             return result;
         case Expires:
             // Display either date of expiry or 'Never'
-            result = entry->timeInfo().expires() ? entry->timeInfo().expiryTime().toLocalTime().toString(EntryModel::DateFormat) : tr("Never");
+            result = entry->timeInfo().expires()
+                         ? entry->timeInfo().expiryTime().toLocalTime().toString(EntryModel::DateFormat)
+                         : tr("Never");
             return result;
         case Created:
             result = entry->timeInfo().creationTime().toLocalTime().toString(EntryModel::DateFormat);
@@ -203,18 +212,35 @@ QVariant EntryModel::data(const QModelIndex& index, int role) const
         case Accessed:
             result = entry->timeInfo().lastAccessTime().toLocalTime().toString(EntryModel::DateFormat);
             return result;
-        case Attachments:
+        case Attachments: {
             // Display comma-separated list of attachments
             QList<QString> attachments = entry->attachments()->keys();
-            for (int i = 0; i < attachments.size(); ++i) {
+            for (const auto& attachment : attachments) {
                 if (result.isEmpty()) {
-                    result.append(attachments.at(i));
+                    result.append(attachment);
                     continue;
                 }
-                result.append(QString(", ") + attachments.at(i));
+                result.append(QString(", ") + attachment);
             }
             return result;
         }
+        case Size: {
+            const int unitsSize = 4;
+            QString units[unitsSize] = {"B", "KiB", "MiB", "GiB"};
+            float resultInt = entry->size();
+
+            for (int i = 0; i < unitsSize; i++) {
+                if (resultInt < 1024 || i == unitsSize - 1) {
+                    resultInt = qRound(resultInt * 100) / 100.0;
+                    result = QStringLiteral("%1 %2").arg(QString::number(resultInt), units[i]);
+                    break;
+                }
+                resultInt /= 1024.0;
+            }
+
+            return result;
+        }
+        }
     } else if (role == Qt::UserRole) { // Qt::UserRole is used as sort role, see EntryView::EntryView()
         switch (index.column()) {
         case Username:
@@ -233,7 +259,11 @@ QVariant EntryModel::data(const QModelIndex& index, int role) const
         case Paperclip:
             // Display entries with attachments above those without when
             // sorting ascendingly (and vice versa when sorting descendingly)
-            return entry->attachments()->isEmpty() ? 1 : 0;
+            return !entry->attachments()->isEmpty();
+        case Totp:
+            return entry->hasTotp();
+        case Size:
+            return entry->size();
         default:
             // For all other columns, simply use data provided by Qt::Display-
             // Role for sorting
@@ -243,17 +273,19 @@ QVariant EntryModel::data(const QModelIndex& index, int role) const
         switch (index.column()) {
         case ParentGroup:
             if (entry->group()) {
-                return entry->group()->iconScaledPixmap();
+                return entry->group()->iconPixmap();
             }
             break;
         case Title:
-            if (entry->isExpired()) {
-                return databaseIcons()->iconPixmap(DatabaseIcons::ExpiredIconIndex);
-            }
-            return entry->iconScaledPixmap();
+            return entry->iconPixmap();
         case Paperclip:
             if (!entry->attachments()->isEmpty()) {
-                return m_paperClipPixmap;
+                return resources()->icon("paperclip");
+            }
+            break;
+        case Totp:
+            if (entry->hasTotp()) {
+                return resources()->icon("chronometer");
             }
             break;
         }
@@ -264,15 +296,23 @@ QVariant EntryModel::data(const QModelIndex& index, int role) const
         }
         return font;
     } else if (role == Qt::ForegroundRole) {
+        QColor foregroundColor;
+        foregroundColor.setNamedColor(entry->foregroundColor());
         if (entry->hasReferences()) {
             QPalette p;
-            return QVariant(p.color(QPalette::Active, QPalette::Mid));
-        } else if (entry->foregroundColor().isValid()) {
-            return QVariant(entry->foregroundColor());
+            foregroundColor = p.color(QPalette::Current, QPalette::Text);
+            int lightness =
+                qMin(255, qMax(0, foregroundColor.lightness() + (foregroundColor.lightness() < 110 ? 85 : -51)));
+            foregroundColor.setHsl(foregroundColor.hue(), foregroundColor.saturation(), lightness);
+            return QVariant(foregroundColor);
+        } else if (foregroundColor.isValid()) {
+            return QVariant(foregroundColor);
         }
     } else if (role == Qt::BackgroundRole) {
-        if (entry->backgroundColor().isValid()) {
-            return QVariant(entry->backgroundColor());
+        QColor backgroundColor;
+        backgroundColor.setNamedColor(entry->backgroundColor());
+        if (backgroundColor.isValid()) {
+            return QVariant(backgroundColor);
         }
     } else if (role == Qt::TextAlignmentRole) {
         if (index.column() == Paperclip) {
@@ -311,19 +351,56 @@ QVariant EntryModel::headerData(int section, Qt::Orientation orientation, int ro
             return tr("Accessed");
         case Attachments:
             return tr("Attachments");
+        case Size:
+            return tr("Size");
         }
+
     } else if (role == Qt::DecorationRole) {
-        if (section == Paperclip) {
-            return m_paperClipPixmap;
+        switch (section) {
+        case Paperclip:
+            return resources()->icon("paperclip");
+        case Totp:
+            return resources()->icon("chronometer");
+        }
+    } else if (role == Qt::ToolTipRole) {
+        switch (section) {
+        case ParentGroup:
+            return tr("Group name");
+        case Title:
+            return tr("Entry title");
+        case Username:
+            return tr("Username");
+        case Password:
+            return tr("Password");
+        case Url:
+            return tr("URL");
+        case Notes:
+            return tr("Entry notes");
+        case Expires:
+            return tr("Entry expires at");
+        case Created:
+            return tr("Creation date");
+        case Modified:
+            return tr("Last modification date");
+        case Accessed:
+            return tr("Last access date");
+        case Attachments:
+            return tr("Attached files");
+        case Size:
+            return tr("Entry size");
+        case Paperclip:
+            return tr("Has attachments");
+        case Totp:
+            return tr("Has TOTP one-time password");
         }
     }
 
-    return QVariant();
+    return {};
 }
 
 Qt::DropActions EntryModel::supportedDropActions() const
 {
-    return 0;
+    return Qt::IgnoreAction;
 }
 
 Qt::DropActions EntryModel::supportedDragActions() const
@@ -419,14 +496,45 @@ void EntryModel::entryRemoved()
     if (m_group) {
         m_entries = m_group->entries();
     }
-
     endRemoveRows();
 }
 
+void EntryModel::entryAboutToMoveUp(int row)
+{
+    beginMoveRows(QModelIndex(), row, row, QModelIndex(), row - 1);
+    if (m_group) {
+        m_entries.move(row, row - 1);
+    }
+}
+
+void EntryModel::entryMovedUp()
+{
+    if (m_group) {
+        m_entries = m_group->entries();
+    }
+    endMoveRows();
+}
+
+void EntryModel::entryAboutToMoveDown(int row)
+{
+    beginMoveRows(QModelIndex(), row, row, QModelIndex(), row + 2);
+    if (m_group) {
+        m_entries.move(row, row + 1);
+    }
+}
+
+void EntryModel::entryMovedDown()
+{
+    if (m_group) {
+        m_entries = m_group->entries();
+    }
+    endMoveRows();
+}
+
 void EntryModel::entryDataChanged(Entry* entry)
 {
     int row = m_entries.indexOf(entry);
-    emit dataChanged(index(row, 0), index(row, columnCount()-1));
+    emit dataChanged(index(row, 0), index(row, columnCount() - 1));
 }
 
 void EntryModel::severConnections()
@@ -446,6 +554,10 @@ void EntryModel::makeConnections(const Group* group)
     connect(group, SIGNAL(entryAdded(Entry*)), SLOT(entryAdded(Entry*)));
     connect(group, SIGNAL(entryAboutToRemove(Entry*)), SLOT(entryAboutToRemove(Entry*)));
     connect(group, SIGNAL(entryRemoved(Entry*)), SLOT(entryRemoved()));
+    connect(group, SIGNAL(entryAboutToMoveUp(int)), SLOT(entryAboutToMoveUp(int)));
+    connect(group, SIGNAL(entryMovedUp()), SLOT(entryMovedUp()));
+    connect(group, SIGNAL(entryAboutToMoveDown(int)), SLOT(entryAboutToMoveDown(int)));
+    connect(group, SIGNAL(entryMovedDown()), SLOT(entryMovedDown()));
     connect(group, SIGNAL(entryDataChanged(Entry*)), SLOT(entryDataChanged(Entry*)));
 }
 
@@ -460,9 +572,10 @@ bool EntryModel::isUsernamesHidden() const
 /**
  * Set state of 'Hide Usernames' setting and signal change
  */
-void EntryModel::setUsernamesHidden(const bool hide)
+void EntryModel::setUsernamesHidden(bool hide)
 {
     m_hideUsernames = hide;
+    emit dataChanged(index(0, 0), index(rowCount() - 1, columnCount() - 1));
     emit usernamesHiddenChanged();
 }
 
@@ -477,29 +590,9 @@ bool EntryModel::isPasswordsHidden() const
 /**
  * Set state of 'Hide Passwords' setting and signal change
  */
-void EntryModel::setPasswordsHidden(const bool hide)
+void EntryModel::setPasswordsHidden(bool hide)
 {
     m_hidePasswords = hide;
+    emit dataChanged(index(0, 0), index(rowCount() - 1, columnCount() - 1));
     emit passwordsHiddenChanged();
 }
-
-/**
- * Toggle state of 'Hide Usernames' setting
- */
-void EntryModel::toggleUsernamesHidden(const bool hide)
-{
-    setUsernamesHidden(hide);
-}
-
-/**
- * Toggle state of 'Hide Passwords' setting
- */
-void EntryModel::togglePasswordsHidden(const bool hide)
-{
-    setPasswordsHidden(hide);
-}
-
-void EntryModel::setPaperClipPixmap(const QPixmap& paperclip)
-{
-    m_paperClipPixmap = paperclip;
-}
